/* Automatically generated by version.sh, do not manually edit! */
#ifndef AVUTIL_FFVERSION_H
#define AVUTIL_FFVERSION_H
#define FFMPEG_VERSION "git-2026-03-16-f550b3be13"
#endif /* AVUTIL_FFVERSION_H */
